;
; << Tiny Prolog Interpreter in Scheme >>
;
;      Coded by Atsushi Moriwaki
;

; assert!
;
; (assert! <predicate> <clause_1> ... <clause_n>)
; ==> (define-clause '(<predicate> <clause1> ... <clause_n>))
;
; Example of usage
;
; (assert! (append () ?x ?x))
; (assert! (append (?a . ?x) ?y (?a . ?z)) (append ?x ?y ?z))

(macro assert! (lambda (l) (list 'define-clause (list 'quote (cdr l)))))

(define (define-clause clause)
  (set! clause (change-variable clause))
  (let ((def (get (caar clause) 'prolog-clauses)))
    (if def
        (nconc def (list clause))
        (put (caar clause) 'prolog-clauses (list clause)))
    (display (caar clause))
    (display " is added to data base.")
    (newline)))

; query?
;
; (query? <query_1> ... <query_n>)
; ==> (query-process '(<query_1> ... <query_n>))
;
; Example of usage
;
; (query? (append (a b) (c d) ?x) (print ?x))

(macro query? (lambda (l) (list 'query-process (list 'quote (cdr l)))))

(define (query-process query)
  (if (prove (change-variable query) '(*dummy*))
      (display "Satisfied.")
      (display "Not satisfied."))
  (newline))


;
; main function of this Prolog interpreter
;

(define (prove query subst)
  (define new-var-list '())
  (define (try-clause clgoal g subst)
    (set! new-var-list '())
    (unify (rename-var clgoal) g subst))
  (define (prove-rest subgoal rest-of-goals subst)
    (prove (nconc (rename-var subgoal) rest-of-goals) subst))
  (define (rename-var form)
    (cond ((var? form)
            (let ((v (assq form new-var-list)))
              (if v
                  (cdr v)
                  (let ((new-var (vector 'prolog-variable
                                          #f
                                          (var-name form))))
                       (set! new-var-list
                             (cons (cons form new-var)
                                   new-var-list))
                       new-var))))
          
          ((pair? form) (cons (rename-var (car form))
                              (rename-var (cdr form))))
          (else form)))

  (cond ((null? query) subst)
        ((var? (car query))
          (prove (cons (fetch-value (car query) subst) (cdr query)) subst))
        
        ((pair? (car query))
          (let ((clauses (get (caar query) 'prolog-clauses)))
            (if (and clauses (pair? clauses))
                (let clauses-loop ((cls clauses))
                     (if (null? cls)
                         #f
                        (let ((new-subst (try-clause (caar cls)
                                                     (car query)
                                                     subst)))
                             (cond ((and new-subst
                                         (prove-rest (cdar cls)
                                                     (cdr query)
                                                     new-subst)))
                                    (else 
                                      (clauses-loop (cdr cls)))))))
                (let ((prim-code (get (caar query) 'prolog-primitive)))
                  (cond (prim-code
                          (let ((new-subst (prim-code (cdar query) subst)))
                            (if new-subst
                                (prove (cdr query) new-subst)
                                #f)))
                        (else #f))))))
        (else (prove (cdr query) subst))))

(define (unify x y subst)
  (define (uni x y subst)
    (cond ((var? x) (bind-var! x y) (cons x subst))
          ((var? y) (bind-var! y x) (cons y subst))
          ((and (pair? x) (pair? y))
            (let ((new-subst (uni (fetch-value (car x) subst)
                                  (fetch-value (car y) subst)
                                  subst)))
               (if new-subst
                   (uni (fetch-value (cdr x) new-subst)
                        (fetch-value (cdr y) new-subst)
                        new-subst)
                   #f)))
          (else (if (equal? x y) subst #f))))
  (uni (fetch-value x subst) (fetch-value y subst) subst))

(define (fetch-value x subst)
  (cond ((var? x)
          (if (memq x subst)
              (fetch-value (bind-value x) subst)
              x))
        ((pair? x) (cons (fetch-value (car x) subst)
                         (fetch-value (cdr x) subst)))
        (else x)))

(define (change-variable form)
  (letrec ((new-var-list '())
           (variable? 
             (lambda (exp)
               (and (symbol? exp)
                    (char=? #\? (string-ref (symbol->string exp) 0)))))
           (iter-change-variable
             (lambda (form)
               (cond ((variable? form)
                       (let ((v (assq form new-var-list)))
                         (if v
                             (cdr v)
                             (let ((new-var
                                    (vector 'prolog-variable #f form)))
                                  (set! new-var-list
                                        (cons (cons form new-var)
                                              new-var-list))
                                  new-var))))
                     ((pair? form) (cons (iter-change-variable (car form))
                                         (iter-change-variable (cdr form))))
                     (else form)))))
          (iter-change-variable form)))

(define (var? exp)
  (and (vector? exp) (eq? (vector-ref exp 0) 'prolog-variable)))

(define (var-name var) (vector-ref var 2))

(define (bind-var! var a) (vector-set! var 1 a))

(define (bind-value var) (vector-ref var 1))

(define (nconc l value)
  (letrec ((get-list-tail
             (lambda (l)
               (if (pair? (cdr l)) (get-list-tail (cdr l)) l))))
    (if (pair? l)
        (begin (set-cdr! (get-list-tail l) value) l)
        value)))

(define (get-print-form form)
  (cond ((var? form) (var-name form))
        ((pair? form) (cons (get-print-form (car form))
                            (get-print-form (cdr form))))
        (else form)))

(define (primitive-print form)
  (write (get-print-form form))
  (newline))

;
; Protable `get' and `put'
;
(define -*-property-list-*- '())

(define (put sym name prop)
  (let ((p-list (assq sym -*-property-list-*-)))
    (if p-list
        (let ((p-pair (assq name (cdr p-list))))
          (if p-pair
              (set-cdr! p-pair prop)
              (nconc p-list (cons (cons name prop) '())))
          #t)
        (begin
          (set! -*-property-list-*- (cons (list sym (cons name prop))
                                          -*-property-list-*-))
          #t))))

(define (get sym name)
  (let ((p-list (assq sym -*-property-list-*-)))
    (if p-list
        (let ((p-pair (assq name (cdr p-list))))
          (if p-pair (cdr p-pair) #f))
       #f)))

; define-primitive-procedure
; See sample primitive definition. 

(define (define-primitive-procedure prim-name proc)
  (put prim-name 'prolog-primitive
    (lambda (form subst)
      (apply proc
             (map (lambda (x) (fetch-value x subst))
                  form))
      subst))
  (display "primitive procedure ")
  (display prim-name)
  (display " is defined.")
  (newline))

; define-primitive-function
; See sample primitive definition.

(define (define-primitive-function prim-name proc)
  (define (get-list-bottom l)
    (if (pair? l)
        (if (pair? (cdr l))
            (get-list-bottom (cdr l))
            (car l))
        #f))
  (define (delete-bottom-from-list l)
    (if (pair? l)
        (if (pair? (cdr l))
            (cons (car l) (delete-bottom-from-list (cdr l)))
            '())
        #f))
  (put prim-name 'prolog-primitive
    (lambda (form subst)
      (unify (apply proc
                    (map (lambda (x) (fetch-value x subst))
                         (delete-bottom-from-list form)))
             (get-list-bottom form)
             subst)))
  (display "primitive function ")
  (display prim-name)
  (display " is defined.")
  (newline))

; define-primitive-boolean
; See sample primitive definition.

(define (define-primitive-boolean prim-name proc)
  (put prim-name 'prolog-primitive
    (lambda (form subst)
      (if (apply proc
                 (map (lambda (x) (fetch-value x subst))
                      form))
          subst
          #f)))
  (display "primitive boolean type function ")
  (display prim-name)
  (display " is defined.")
  (newline))

;
; sample primitive definition
;
(define-primitive-procedure 'print primitive-print)
(define-primitive-procedure 'print1 write)
(define-primitive-procedure 'princ display)
(define-primitive-procedure 'new-line newline)
(define-primitive-function '+ +)
(define-primitive-function '- -)
(define-primitive-function '* *)
(define-primitive-function '/ /)
(define-primitive-function 'quotient quotient)
(define-primitive-function 'remainder remainder)
(define-primitive-boolean '> >)
(define-primitive-boolean '< <)
(define-primitive-boolean '>= >=)
(define-primitive-boolean '<= <=)

(put 'and 'prolog-primitive
  (lambda (form subst)
    (prove form subst)))

(put '= 'prolog-primitive
  (lambda (form subst)
    (unify (car form) (cadr form) subst)))

(put 'not 'prolog-primitive
  (lambda (form subst)
    (if (prove form subst)
        #f
        subst)))

(put 'if 'prolog-primitive
  (lambda (form subst)
    (if (prove (list (car form)) subst)
        (prove (list (cadr form)) subst)
        (prove (list (caddr form)) subst))))

;
; sample predicate
;

(assert! (or ?p . ?q) ?p)
(assert! (or ?p . ?q) (or . ?q))

(assert! (false) (= 1 2))

(assert! (true))

(assert! (member-of ?x (?x . ?y)))
(assert! (member-of ?x (?y . ?z)) (member-of ?x ?z))

(assert! (append () ?x ?x))
(assert! (append (?a . ?x) ?y (?a . ?z)) (append ?x ?y ?z))

;
; test
;

(query? (member-of 9 (1 2 3 4 5)))

(query? (append ?x ?y (0 1 2 3 4 5 6 7 8 9))
        (print ?x)
        (print ?y)
        (false))

(assert! (fact 0 1))
(assert! (fact ?n ?f)
         (- ?n 1 ?n1)
         (fact ?n1 ?f1)
         (* ?n ?f1 ?f))
(query? (fact 10 ?x) (print ?x))

(assert! (hanoi 1 ?from ?to ?via) (move 1 ?from ?to))
(assert! (hanoi ?n ?from ?to ?via)
         (- ?n 1 ?n1)
         (hanoi ?n1 ?from ?via ?to)
         (move ?n ?from ?to)
         (hanoi ?n1 ?via ?to ?from))
(assert! (move ?n ?from ?to) (printlist (Move ?n from ?from to ?to)))
(assert! (printlist ()) (princ "!") (new-line))
(assert! (printlist (?x . ?y)) (print1 ?x) (princ " ") (printlist ?y))
(query? (hanoi 3 a b c))
