;;;; AM-Scheme utility procedures
;;;
;;;  Written by Akira KIDA,  <SDI00379@niftyserve.or.jp>
;;;
;;;  This software is placed in the public domain
;;;

;;
;; The following are useful for trace, pretty-print and ed.
;;
(define (symbol->function name)
    (let ((code (eval name)))
        (if (closure? code)
            (get-closure-code code)
            #f)))

(define (macro->function name)
    (let ((code (eval name)))
        (if (macro? code)
            (get-closure-code code)
            #f)))     

;;; print n spaces
(define (spaces n . port)
  (if (positive? n)
    (begin
      (display " " . port)
      (spaces (- n 1) . port))))

;;; get definition of a procedure or a macro
;;; yes, this name comes from one of InterLisp. :-)
(define (getd symbol)
  (let ((def (symbol->function symbol)))
    (if def
      (cons
        'define
        (cons
          (cons symbol (car (cdr def)))
          (cdr (cdr def))))
      (begin
        (set! def (macro->function symbol))
        (if def
          (cons 'macro (list symbol def))
          (list 'error "Not a S-Expression procedure:"
            (list 'quote symbol)))))))



;;;;    A Simple Trace package for Mini-Scheme
;;
;;      Date written    15-Dec-1989 by Akira Kida (akida)
;;
(define *trace-indent* 0)       ;  holds indent level during trace

;;; Change -*-reset-patch-*-
(set! -*-reset-patch-*-
  (let ((old-reset-patch -*-reset-patch-*-))
    (lambda () (set! *trace-indent* 0) (old-reset-patch))))

;;;  run-time support procedure for trace
(define (@trace-procedure-call name args)
  (let ((proc (get name 'expr))
        (form (cons name args)))
    (if (not proc) (error "Cannot trace " name))
    (spaces *trace-indent*)
    (set! *trace-indent* (+ *trace-indent* 2))
    (display form)
    (newline)
    (let ((result (apply proc args)))
      (set! *trace-indent* (- *trace-indent* 2))
      (if (negative? *trace-indent*) (set! *trace-indent* 0))
      (spaces *trace-indent*)
      (display form)
      (display " ==> ")
      (display result)
      (newline)
      result)))

;;; set/unset trace patch for a procedure
(define (set-trace name set)
  (if (not (procedure? (eval name)))
    (error "Cannot set trace :" name))
  (if (macro->function name)
    (error "Cannot set trace on macro :" name))
  (let ((saved-proc (get name 'expr)))
    (if set
      (if (or (not saved-proc)
              (not
                (eq? (car (car (cdr (cdr (getd name)))))
                     '@trace-procedure-call)))
        (begin
          (put name 'expr (eval name))
          (eval
            (list 'set! name
              (list 'lambda 'args
                (list
                  '@trace-procedure-call
                  (list 'quote name)
                  'args))))))
      (if (and saved-proc
               (eq? (car (car (cdr (cdr (getd name)))))
                    '@trace-procedure-call))
        (begin
          (put name 'expr '())
          (eval (list 'set! name saved-proc))))))
  name)

;;; set/unset trace patch for procedures
(define (set-trace-procs symbols set)
  (if (pair? symbols)
    (map (lambda (f) (set-trace f set)) symbols)
    (set-trace symbols set)))

;;; Trace package user interface
;;;
;;; (trace sym1 sym2 ...)       ; trace enable sym1 sym2 ...
;;; (untrace sym1 sym2 ...)     ; trace disable sym1 sym2 ...
;;;
;;; Since these are macros, arguments need not be quoted.

;;; set trace patch for prcedures
(macro trace
  (lambda (trace-macro)
    (set! *trace-indent* 0)
    (list 'set-trace-procs (list 'quote (cdr trace-macro)) #t)))

;;; unset trace patch for procedures
(macro untrace
  (lambda (untrace-macro)
    (set! *trace-indent* 0)
    (list 'set-trace-procs (list 'quote (cdr untrace-macro)) #f)))



;;;;    A Tiny Pretty Printer for AM-Scheme
;;;
;;;     Date written    28-Nov-1989 by Akira Kida (for Mini-Scheme)
;;;     Date revised    30-Nov-1989 by Akira Kida
;;;     Date revised    14-Dec-1989 by Akira Kida
;;;     Date revised    24-Jan-1990 by Atsushi Moriwaki (for SECDR-Scheme)
;;;     Date revised    18-Mar-1992 by Akira Kida (for AM-Scheme)
;;;

; Columns of display device.
(define *pp-display-width* 80)

; Margin of display-width
(define *pp-display-margin* 0.9)

; Number of elements will possibly be displayed in one line.
(define *pp-display-elements* 16)


;; pretty print main procedure
;;
;; (pretty-print a-list)        -- pretty print ``a-list''
;; (pretty-print a-list port)   --  output directed to ``port''
;;
(define (pretty-print a-list . port)
  ; List of procedures which need exceptional handling.
  ; Structure or each element in the list is
  ; 
  ;     (name . special-indentation)
  ;
  ;     where name is a symbol and
  ;     special-indentaion is an integer.
  ;
  ; #1 Standard format, non special case.
  ;  (proc
  ;    arg1
  ;    arg2
  ;    arg3)
  ;
  ; #2 Format for special-indentation == 0
  ;  (proc arg1
  ;    arg2
  ;    arg3)
  ;
  ; #3 Format for special-indentaion == 1
  ;  (proc arg1
  ;        arg2
  ;        arg3)
  ;
  ; #4 Format for let, do == 2
  ;  (let ((x .....)
  ;        (y .....))       ; <-- newline here is mandatory
  ;    <....>
  ;    <....>)
  ;
  (define exception
    '((lambda . 0) (if . 0) (and . 1)
      (or . 1) (let . 2) (case . 0)
      (define . 0) (macro . 0)
      (map . 0) (apply . 0)
      (eq? . 1) (eqv? . 1) (set! . 0)
      (let* . 2) (letrec . 2)
      (* . 1) (/ . 1) (+ . 1) (- . 1)
      (= . 1) (< . 1) (> . 1) (<= . 1) (>= . 1)
      (do . 2)
      (call-with-input-file . 0) (call-with-output-file . 0)))
  ; special quasi-procedure
  (define special
    '((quote 1 . "'") (quasiquote 1 . "`")
      (unquote 2 . ",") (unquote-splicing 2 . ",@")))
  (define pp-margin (round (* *pp-display-width* *pp-display-margin*)))
  ; check wheter the number of elements exceeds n or not.
  (define (less-than-n-elements? a-list n)
    ; count elements in a-list at most (n+1)
    (define (up-to-nth a-list n c)
      (cond
        ((null? a-list) c)
        ((pair? a-list)
          (set! c (up-to-nth (car a-list) n c))
          (if (< n c)
            c
            (up-to-nth (cdr a-list) n c)))
        (else (+ c 1))))
    (< (up-to-nth a-list n 0) n))
  ; check if the length is fit within n colums or not.
  (define (fit-in-n-width? a-list n)
    (< (list-width a-list) n))
  ; indent and pretty-print
  (define (do-pp a-list col)
    (spaces col . port)
    (pp-list a-list col 2))
  ;; main logic.
  (define (pp-list a-list col step)
    (cond
      ((atom? a-list) (write a-list . port))     ; atom
      ((and (assq (car a-list) special)
            (pair? (cdr a-list))
            (null? (cddr a-list)))   ; check for proper quote etc.
        (let ((s (assq (car a-list) special)))
          (display (cddr s) . port)             ; display using abbrev.
          (pp-list
            (cadr a-list)
            (+ col (- (list-width (cddr s)) 2))
            (cadr s))))
      (else
        (let* ((sym (car a-list))
               (ex-col (assq sym exception)))
	  (if (and (not (and ex-col
			     (= (cdr ex-col) 2)
			     (pair? (cdr a-list))
			     (not (null? (cddr a-list)))))
		   (less-than-n-elements? a-list *pp-display-elements*)
		   (fit-in-n-width? a-list (- pp-margin col)))
	      ;; short list or type != 2
	      (begin
		(display "(" . port)
		(do-pp (car a-list) 0)
		(pp-args #f (cdr a-list) 1))
	      ;; long list or type == 2
	      (if (pair? ex-col)              ; check for exception.,
		(case (cdr ex-col)
		  ((0 1)
		    (display "(" . port)
		    (write sym . port)
		    (display " " . port)
		    (pp-list (cadr a-list) (+ col 2 (list-width sym)) 2)
		    (pp-args
		     #t
		     (cdr (cdr a-list))
		     (+ col
			2
			(if (zero? (cdr ex-col)) 0 (list-width sym)))))
		  ((2)
		    (display "(" . port)
		    (write sym  . port)
		    (display " " . port)
		    (if (symbol? (cadr a-list))
			(begin ; named let
			  (write (cadr a-list) . port)
			  (display " " . port)
			  (pp-list
			   (caddr a-list)
			   (+ col
			      3
			      (list-width sym)
			      (list-width (cadr a-list)))
			   1)
			  (pp-args #t (cdddr a-list) (+ col 2)))
			(begin ; usual let
			  (pp-list (cadr a-list) (+ col 2 (list-width sym)) 1)
			  (pp-args #t (cddr a-list) (+ col 2)))))
		  (else
		   (error "Illegal exception")))
		(begin                        ; normal case.
		  (display "(" . port)
		  (pp-list (car a-list) (+ col 1) 2)
		  (pp-args #t (cdr a-list) (+ col step)))))))))
  ;; display arguments
  (define (pp-args nl a-list col)
    (cond
      ((null? a-list) (display ")" . port))
      ((pair? a-list)
	(if nl (newline . port))
        (do-pp (car a-list) col)
        (pp-args nl (cdr a-list) col))
      (else
        (display " . " . port)
        (write a-list . port)
        (display ")" . port))))
  (do-pp a-list 0)
  (newline . port))


;;; pretty print procedure(s)/macro(s).
;;;   (pretty 'a-symbol)        ; pretty print a procedure or macro
;;;   (pretty '(sym1 sym2 ...)) ; pretty print procedures and/or macros
(define (pretty symbols . port)
  (if (pair? symbols)
    (for-each
      (lambda (x) (pretty-print (getd x) . port) (newline . port))
      symbols)
    (pretty-print (getd symbols) . port)))

(define (pretty-file symbols file)
  (call-with-output-file file
    (lambda (p) (pretty symbols p))))

;;; pretty print user-interface
;;;
;;; usage:
;;;  (pp sym1 sym2 ...)  ; obtain procedure/macro definitions in sequence
;;;
;;; Note: pp never evaluate its argument, so you do not have to specify
;;;       (pp 'proc-name). Use (pp proc-name) insted.
;;;
(macro pp (lambda (pp-macro)
  (list 'pretty (list 'quote (cdr pp-macro)))))



;;;;    A Very Tiny Lisp Editor for Mini-Scheme
;;;
;;;     Date written    15-Dec-1989 by Akira Kida
;;;     Date added      18-Dec-1989 by Atsushi Moriwaki
;;;
;;; Synopsis:
;;;     (ed symbol)     --- symbol is a unquoted procedure name.
;;;                         You can edit both procedures defined in
;;;                         s-expression and macros.
;;;                         If symbol does not represent a s-expression
;;;                         defined procedure nor a macro, an error
;;;                         is reported.
;;;
;;; Command:
;;;     a --    Change current sublist to CAR of former one.
;;;     d --    Change current sublist to CDR of former one.
;;;     l --    Print outline of current sublist.
;;;     p --    Pretty print current sublist.
;;;     r --    Replace current sublist to one typed follow.
;;;     i --    Insert one typed follow to CAR.
;;;     x --    Delete CAR of current sublist.
;;;     c --    Copy current buffer to CAR.
;;;     v --    Yank current sublist to buffer.
;;;     e --    Edit buffer.
;;;     b --    Show buffer.
;;;     u --    Change current sublist up one level in list.
;;;     ! --    Exit to toplevel.
;;;
(macro ed
  (lambda (ed-macro)
    (define (edit-procedure name)
      (let ((def (getd name)))
        (if (null? def)
          (error
            "Cannot edit non S-Expression procedure/macro"
            name))
        (set! def (edit-list def))
        (let ((ans #f))
          (define (confirm)
            (let ((ans #f))
              (display "A)bandan, E)valuate, C)ontinue : ")
              (set! ans (read))
              (if (not (memq ans '(a A e E c C)))
                (confirm)
                ans)))
          (define (edit-again)
            (let ((ans (confirm)))
              (if (memq ans '(c C))
                (begin (set! def (edit-list def)) (edit-again))
                ans)))
          (set! ans (edit-again))
          (if (memq ans '(e E))
            (if (memq (car def) '(define macro))
              def
              (list
                'error
                '"invalid definition syntax."
                (list 'quote (car def))))
            ''aborted))))
    (let* ((name (car (cdr ed-macro)))
            (evalform (edit-procedure name)))
      evalform)))

(define (print-outline lis howmany)
  (define (prt-ol lis done)
    (cond
      ((<= howmany done)
        (display "&")
        (+ done 1))
      ((pair? lis)
        (display "(")
        (prt-ol-tail (cdr lis) (prt-ol (car lis) done)))
      (else (write lis) (+ done 1))))
  (define (prt-ol-tail lis done)
    (cond
      ((null? lis) (display ")") done)
      ((pair? lis)
        (display " ")
        (prt-ol-tail (cdr lis) (prt-ol (car lis) done)))
      (else
        (display " . ")
        (let ((result (prt-ol lis done)))
          (display ")")
          result))))
  (prt-ol lis 0)
  (newline))

(define (edit-list lis)
  (let ((exit-flag #f) (buffer '()))
    (define (ed-list lis level)
      (display "a,d,l,p,r,i,x,c,v,e,b,u,?,! (")
      (if (zero? level) (display "top") (display level))
      (display "): ")
      (let ((cmd (read)))
        (case cmd
          ((a A)
            (if (pair? lis)
              (set! lis
                (cons (ed-list (car lis) (+ level 1)) (cdr lis)))
              (begin (write "cannot execute CAR command!\n"))))
          ((d D)
            (if (pair? lis)
              (set! lis
                (cons (car lis) (ed-list (cdr lis) (+ level 1))))
              (begin (write "cannot execute CDR command!\n"))))
          ((l L) (print-outline lis 5))
          ((p P) (pretty-print lis))
          ((r R) (display "? ") (set! lis (read)))
          ((i I) (display "? ") (set! lis (cons (read) lis)))
          ((x X) (set! buffer (car lis)) (set! lis (cdr lis)))
          ((c C) (set! lis (cons buffer lis)))
          ((v V) (set! buffer lis))
          ((e E) (display "? ") (set! buffer (read)))
          ((b B) (write buffer) (newline))
          ((?)
            (display " ----- list editor command ----\n")
            (display " a -- Change current sublist to CAR\n")
            (display " d -- Change current sublist to CDR\n")
            (display " l -- Print outline of current sublist\n")
            (display " p -- Pretty print current sublist\n")
            (display " r -- Replace current sublist to one typed follow\n")
            (display " i -- Insert one typed follow to CAR\n")
            (display " x -- Delete CAR of current sublist\n")
            (display " c -- Copy current buffer to CAR\n")
            (display " v -- Yank current sublist to buffer\n")
            (display " e -- Edit buffer\n")
            (display " b -- Show buffer\n")
            (display " u -- Exit current sublist (up one level)\n")
            (display " ! -- Exit to toplevel\n"))
          ((u U) #t)
          ((!) (set! exit-flag #t))
          (else
            (display "Command error. use ? to help.\n")))
        (if (and exit-flag (= level 0))
          (set! exit-flag #f))
        (if (or exit-flag (memq cmd '(u U)))
          lis
          (ed-list lis level))))
    (ed-list lis 0)))

